#include <msp/gl/extensions/arb_texture_float.h>
#include <msp/gl/extensions/ext_bgra.h>
#include <msp/strings/format.h>
#include "pixelformat.h"

using namespace std;

namespace Msp {
namespace GL {

void operator>>(const LexicalConverter &conv, PixelFormat &fmt)
{
	if(conv.get()=="STENCIL_INDEX")
		fmt = STENCIL_INDEX;
	else if(conv.get()=="DEPTH_COMPONENT")
		fmt = DEPTH_COMPONENT;
	else if(conv.get()=="RED")
		fmt = RED;
	else if(conv.get()=="RG")
		fmt = RG;
	else if(conv.get()=="RGB")
		fmt = RGB;
	else if(conv.get()=="RGBA")
		fmt = RGBA;
	else if(conv.get()=="BGR")
		fmt = BGR;
	else if(conv.get()=="BGRA")
		fmt = BGRA;
	else if(conv.get()=="SRGB")
		fmt = SRGB;
	else if(conv.get()=="SRGB_ALPHA")
		fmt = SRGB_ALPHA;
	else if(conv.get()=="LUMINANCE")
		fmt = LUMINANCE;
	else if(conv.get()=="LUMINANCE_ALPHA")
		fmt = LUMINANCE_ALPHA;
	else if(conv.get()=="SLUMINANCE")
		fmt = SLUMINANCE;
	else if(conv.get()=="SLUMINANCE_ALPHA")
		fmt = SLUMINANCE_ALPHA;
	else
		throw lexical_error(format("conversion of '%s' to PixelFormat", conv.get()));
}

PixelFormat pixelformat_from_graphics(Graphics::PixelFormat pf)
{
	switch(pf)
	{
	case Graphics::LUMINANCE: return LUMINANCE;
	case Graphics::LUMINANCE_ALPHA: return LUMINANCE_ALPHA;
	case Graphics::RGB: return RGB;
	case Graphics::RGBX:
	case Graphics::RGBA: return RGBA;
	case Graphics::BGR: return BGR;
	case Graphics::BGRX:
	case Graphics::BGRA: return BGRA;
	default: throw invalid_argument("pixelformat_from_graphics");
	}
}

PixelFormat storage_pixelformat_from_graphics(Graphics::PixelFormat pf, bool srgb)
{
	PixelFormat result;
	switch(pf)
	{
	case Graphics::RGBX:
	case Graphics::BGR:
	case Graphics::BGRX: result = RGB; break;
	case Graphics::BGRA: result = RGBA; break;
	default: result = pixelformat_from_graphics(pf);
	}

	if(srgb)
		return get_srgb_pixelformat(result);
	else
		return result;
}

PixelFormat get_base_pixelformat(PixelFormat pf)
{
	PixelFormat unsized = get_unsized_pixelformat(pf);
	switch(unsized)
	{
	case SRGB: return RGB;
	case SRGB_ALPHA: return RGBA;
	case SLUMINANCE: return LUMINANCE;
	case SLUMINANCE_ALPHA: return LUMINANCE_ALPHA;
	default: return unsized;
	}
}

PixelFormat get_unsized_pixelformat(PixelFormat pf)
{
	switch(pf)
	{
	case R8:
	case R16F:
	case R32F: return RED;
	case RG8:
	case RG16F:
	case RG32F: return RG;
	case RGB8:
	case RGB16F:
	case RGB32F: return RGB;
	case SRGB8: return SRGB;
	case RGBA8:
	case RGBA16F:
	case RGBA32F: return RGBA;
	case SRGB8_ALPHA8: return SRGB_ALPHA;
	case LUMINANCE8: return LUMINANCE;
	case SLUMINANCE8: return SLUMINANCE;
	case LUMINANCE_ALPHA8: return LUMINANCE_ALPHA;
	case SLUMINANCE8_ALPHA8: return SLUMINANCE_ALPHA;
	case DEPTH_COMPONENT16:
	case DEPTH_COMPONENT24:
	case DEPTH_COMPONENT32: return DEPTH_COMPONENT;
	default: return pf;
	}
}

PixelFormat get_sized_pixelformat(PixelFormat pf, unsigned size)
{
	if(!size || size>4)
		throw invalid_argument("get_sized_pixelformat");

	switch(pf)
	{
	case SRGB:
	case SRGB_ALPHA: break;
	case SRGB8: pf = SRGB; break;
	case SRGB8_ALPHA8: pf = SRGB_ALPHA; break;
	default: pf = get_base_pixelformat(pf);
	}

	switch(size)
	{
	case 1:
		switch(pf)
		{
		case RED: return R8;
		case RG: return RG8;
		case RGB: return RGB8;
		case RGBA: return RGBA8;
		case SRGB: return SRGB8;
		case SRGB_ALPHA: return SRGB8_ALPHA8;
		case LUMINANCE: return LUMINANCE8;
		case SLUMINANCE: return SLUMINANCE8;
		case LUMINANCE_ALPHA: return LUMINANCE8_ALPHA8;
		case SLUMINANCE_ALPHA: return SLUMINANCE8_ALPHA8;
		default: throw invalid_argument("get_sized_pixelformat");
		}
	case 2:
		switch(pf)
		{
		case RED: return R16F;
		case RG: return RG16F;
		case RGB: return RGB16F;
		case RGBA: return RGBA16F;
		case DEPTH_COMPONENT: return DEPTH_COMPONENT16;
		default: throw invalid_argument("get_sized_pixelformat");
		}
	case 3:
		if(pf==DEPTH_COMPONENT)
			return DEPTH_COMPONENT24;
		else
			throw invalid_argument("get_sized_pixelformat");
	case 4:
		switch(pf)
		{
		case RED: return R32F;
		case RG: return RG32F;
		case RGB: return RGB32F;
		case RGBA: return RGBA32F;
		case DEPTH_COMPONENT: return DEPTH_COMPONENT32;
		default: throw invalid_argument("get_sized_pixelformat");
		}
	default:
		throw invalid_argument("get_sized_pixelformat");
	}
}

PixelFormat get_srgb_pixelformat(PixelFormat pf)
{
	switch(pf)
	{
	case RGB: return SRGB;
	case RGBA: return SRGB_ALPHA;
	case RGB8: return SRGB8;
	case RGBA8: return SRGB8_ALPHA8;
	case LUMINANCE: return SLUMINANCE;
	case LUMINANCE8: return SLUMINANCE8;
	case LUMINANCE_ALPHA: return SLUMINANCE_ALPHA;
	case LUMINANCE_ALPHA8: return SLUMINANCE8_ALPHA8;
	default: return pf;
	}
}

unsigned get_component_count(PixelFormat pf)
{
	switch(get_base_pixelformat(pf))
	{
	case STENCIL_INDEX:
	case DEPTH_COMPONENT:
	case RED:
	case LUMINANCE:
	case SLUMINANCE:
		return 1;
	case RG:
	case LUMINANCE_ALPHA:
	case SLUMINANCE_ALPHA:
		return 2;
	case RGB:
	case BGR:
		return 3;
	case RGBA:
	case BGRA:
		return 4;
	default:
		throw invalid_argument("get_component_count");
	}
}

unsigned get_component_size(PixelFormat pf)
{
	switch(pf)
	{
	case R8:
	case RG8:
	case RGB8:
	case RGBA8:
	case SRGB8:
	case SRGB8_ALPHA8:
	case LUMINANCE8:
	case LUMINANCE8_ALPHA8:
	case SLUMINANCE8:
	case SLUMINANCE8_ALPHA8:
		return 1;
	case R16F:
	case RG16F:
	case RGB16F:
	case RGBA16F:
	case DEPTH_COMPONENT16:
		return 2;
	case DEPTH_COMPONENT24:
		return 3;
	case R32F:
	case RG32F:
	case RGB32F:
	case RGBA32F:
	case DEPTH_COMPONENT32:
		return 4;
	default:
		return 0;
	}
}

unsigned get_pixel_size(PixelFormat pf)
{
	return get_component_count(pf)*max(get_component_size(pf), 1U);
}

void require_pixelformat(PixelFormat pf)
{
	switch(pf)
	{
	case RED:
	case R8:
	case RG:
	case RG8:
		{ static Require _req(ARB_texture_rg); }
		break;
	case R16F:
	case R32F:
	case RG16F:
	case RG32F:
		{ static Require _req(ARB_texture_rg); }
		{ static Require _req(ARB_texture_float); }
		break;
	case RGB16F:
	case RGB32F:
	case RGBA16F:
	case RGBA32F:
		{ static Require _req(ARB_texture_float); }
		break;
	case SRGB:
	case SRGB8:
	case SRGB_ALPHA:
	case SRGB8_ALPHA8:
	case SLUMINANCE:
	case SLUMINANCE8:
	case SLUMINANCE_ALPHA:
	case SLUMINANCE8_ALPHA8:
		{ static Require _req(EXT_texture_sRGB); }
		break;
	case BGR:
	case BGRA:
		{ static Require _req(EXT_bgra); }
		break;
	case DEPTH_COMPONENT:
	case DEPTH_COMPONENT16:
	case DEPTH_COMPONENT24:
	case DEPTH_COMPONENT32:
		{ static Require _req(ARB_depth_texture); }
		break;
	default:
		break;
	}
}

} // namespace GL
} // namespace Msp
